flsys = LibStub("AceAddon-3.0"):NewAddon("flsys", "AceConsole-3.0", "AceEvent-3.0", "AceComm-3.0")
flsys:SetDefaultModuleState(false)

-------------------------------------------------------------------------------
--- VARIABLES
-------------------------------------------------------------------------------
-- local pointers to db data
local db, raidstocount, min_gp, ep_amt, ep_recur, standard_ilvl, alts_data

-- delay/timer frames
local fra = CreateFrame("Frame", "flsys_TimerFrame")
local delay_fra = CreateFrame("Frame", "flsys_DelayTimerFrame")

local numraidmembers
local rowbuttons = {} --main frame
local maindisplay = 1
local log_rows = {} -- log row buttons
local rosterpage = 0

--- some key tables:
-- db.players
--    player name key maps to table with all their info
--    temp is used in mode 2 to indicate subs
--    alt for alts
--
-- db.sortlist
--    all the same people as db.players, but indexed by number for sorting
--    during raid, will include temps, alts, etc.

-------------------------------------------------------------------------------
--- CONSTANTS
-------------------------------------------------------------------------------

local DEFAULT_STANDARD_ILVL = 553
-- following tables modified from LibGearPoints-1.2 (EPGP)
local EQUIPSLOT_MULTIPLIER_1 = {
  INVTYPE_HEAD = 1,
  INVTYPE_NECK = 0.5,
  INVTYPE_SHOULDER = 0.75,
  INVTYPE_CHEST = 1,
  INVTYPE_ROBE = 1,
  INVTYPE_WAIST = 0.75,
  INVTYPE_LEGS = 1,
  INVTYPE_FEET = 0.75,
  INVTYPE_WRIST = 0.5,
  INVTYPE_HAND = 0.75,
  INVTYPE_FINGER = 0.5,
  INVTYPE_TRINKET = 1,
  INVTYPE_CLOAK = 0.5,
  INVTYPE_WEAPON = 1.5,
  INVTYPE_SHIELD = 1.5,
  INVTYPE_2HWEAPON = 2,
  INVTYPE_WEAPONMAINHAND = 1,
  INVTYPE_WEAPONOFFHAND = 1,
  INVTYPE_HOLDABLE = 0.5, --off-hand frill
  INVTYPE_RANGED = 2, --bow
  INVTYPE_RANGEDRIGHT = 2, --wand, crossbow or gun
}
local EQUIPSLOT_MULTIPLIER_2 = {
  INVTYPE_WEAPON = 0.5, --tanking = .5, other main hand = 1.5
  INVTYPE_SHIELD = 0.5, --tanking shields = 1.5, shaman = .5
  INVTYPE_RANGEDRIGHT = 1.5, --wands = 1.5, xbow/gun = 2
}
local CUSTOM_ITEM_DATA = {
  --T14 normal
  [89234] = { 496, "INVTYPE_HEAD" },
  [89235] = { 496, "INVTYPE_HEAD" },
  [89236] = { 496, "INVTYPE_HEAD" },
  [89237] = { 496, "INVTYPE_CHEST" },
  [89238] = { 496, "INVTYPE_CHEST" },
  [89239] = { 496, "INVTYPE_CHEST" },
  [89240] = { 496, "INVTYPE_HAND" },
  [89241] = { 496, "INVTYPE_HAND" },
  [89242] = { 496, "INVTYPE_HAND" },
  [89243] = { 496, "INVTYPE_LEGS" },
  [89244] = { 496, "INVTYPE_LEGS" },
  [89245] = { 496, "INVTYPE_LEGS" },
  [89246] = { 496, "INVTYPE_SHOULDER" },
  [89247] = { 496, "INVTYPE_SHOULDER" },
  [89248] = { 496, "INVTYPE_SHOULDER" },
  --T14 heroic
  [89249] = { 509, "INVTYPE_CHEST" },
  [89250] = { 509, "INVTYPE_CHEST" },
  [89251] = { 509, "INVTYPE_CHEST" },
  [89252] = { 509, "INVTYPE_LEGS" },
  [89253] = { 509, "INVTYPE_LEGS" },
  [89254] = { 509, "INVTYPE_LEGS" },
  [89255] = { 509, "INVTYPE_HAND" },
  [89256] = { 509, "INVTYPE_HAND" },
  [89257] = { 509, "INVTYPE_HAND" },
  [89258] = { 509, "INVTYPE_HEAD" },
  [89259] = { 509, "INVTYPE_HEAD" },
  [89260] = { 509, "INVTYPE_HEAD" },
  [89261] = { 509, "INVTYPE_SHOULDER" },
  [89262] = { 509, "INVTYPE_SHOULDER" },
  [89263] = { 509, "INVTYPE_SHOULDER" },
  -- T15 normal
  [95573] = { 522, "INVTYPE_SHOULDER" },
  [95583] = { 522, "INVTYPE_SHOULDER" },
  [95578] = { 522, "INVTYPE_SHOULDER" },
  [95572] = { 522, "INVTYPE_LEGS" },
  [95581] = { 522, "INVTYPE_LEGS" },
  [95576] = { 522, "INVTYPE_LEGS" },
  [95571] = { 522, "INVTYPE_HEAD" },
  [95582] = { 522, "INVTYPE_HEAD" },
  [95577] = { 522, "INVTYPE_HEAD" },
  [95570] = { 522, "INVTYPE_HAND" },
  [95580] = { 522, "INVTYPE_HAND" },
  [95575] = { 522, "INVTYPE_HAND" },
  [95569] = { 522, "INVTYPE_CHEST" },
  [95579] = { 522, "INVTYPE_CHEST" },
  [95574] = { 522, "INVTYPE_CHEST" },
  -- T15 heroic
  [96699] = { 535, "INVTYPE_SHOULDER" },
  [96700] = { 535, "INVTYPE_SHOULDER" },
  [96701] = { 535, "INVTYPE_SHOULDER" },
  [96631] = { 535, "INVTYPE_LEGS" },
  [96632] = { 535, "INVTYPE_LEGS" },
  [96633] = { 535, "INVTYPE_LEGS" },
  [96625] = { 535, "INVTYPE_HEAD" },
  [96623] = { 535, "INVTYPE_HEAD" },
  [96624] = { 535, "INVTYPE_HEAD" },
  [96599] = { 535, "INVTYPE_HAND" },
  [96600] = { 535, "INVTYPE_HAND" },
  [96601] = { 535, "INVTYPE_HAND" },
  [96567] = { 535, "INVTYPE_CHEST" },
  [96568] = { 535, "INVTYPE_CHEST" },
  [96566] = { 535, "INVTYPE_CHEST" },
  -- T16 Normal
  [99685] = { 4, 553, "INVTYPE_SHOULDER" },
  [99695] = { 4, 553, "INVTYPE_SHOULDER" },
  [99690] = { 4, 553, "INVTYPE_SHOULDER" },
  [99684] = { 4, 553, "INVTYPE_LEGS" },
  [99693] = { 4, 553, "INVTYPE_LEGS" },
  [99688] = { 4, 553, "INVTYPE_LEGS" },
  [99683] = { 4, 553, "INVTYPE_HEAD" },
  [99694] = { 4, 553, "INVTYPE_HEAD" },
  [99689] = { 4, 553, "INVTYPE_HEAD" },
  [99682] = { 4, 553, "INVTYPE_HAND" },
  [99692] = { 4, 553, "INVTYPE_HAND" },
  [99687] = { 4, 553, "INVTYPE_HAND" },
  [99696] = { 4, 553, "INVTYPE_CHEST" },
  [99691] = { 4, 553, "INVTYPE_CHEST" },
  [99686] = { 4, 553, "INVTYPE_CHEST" },
  -- T16 Normal Essences
  [105857] = { 4, 553, "INVTYPE_HEAD" },
  [105859] = { 4, 553, "INVTYPE_HEAD" },
  [105858] = { 4, 553, "INVTYPE_HEAD" },
  -- T16 Heroic
  [99717] = { 4, 566, "INVTYPE_SHOULDER" },
  [99719] = { 4, 566, "INVTYPE_SHOULDER" },
  [99718] = { 4, 566, "INVTYPE_SHOULDER" },
  [99726] = { 4, 566, "INVTYPE_LEGS" },
  [99713] = { 4, 566, "INVTYPE_LEGS" },
  [99712] = { 4, 566, "INVTYPE_LEGS" },
  [99723] = { 4, 566, "INVTYPE_HEAD" },
  [99725] = { 4, 566, "INVTYPE_HEAD" },
  [99724] = { 4, 566, "INVTYPE_HEAD" },
  [99720] = { 4, 566, "INVTYPE_HAND" },
  [99722] = { 4, 566, "INVTYPE_HAND" },
  [99721] = { 4, 566, "INVTYPE_HAND" },
  [99714] = { 4, 566, "INVTYPE_CHEST" },
  [99716] = { 4, 566, "INVTYPE_CHEST" },
  [99715] = { 4, 566, "INVTYPE_CHEST" },
  -- T16 Heroic Essences
  [105868] = { 4, 566, "INVTYPE_HEAD" },
  [105867] = { 4, 566, "INVTYPE_HEAD" },
  [105866] = { 4, 566, "INVTYPE_HEAD" },
}

-------------------------------------------------------------------------------
--- OPTIONS & DB DEFAULTS
-------------------------------------------------------------------------------
local defaults = {
  profile={
    raidstocount = 15,
    min_gp = 500,
    ep_amt = 250,
    ep_recur = 30,
    players = {},
    alts_data = {},
    dates = {},
    sortlist = {},
    next_award=0,
    lastbroadcast = 0,
    epgplog = {},
    epgp_export = 0, --timestamp
    standings_export = 0, --timestamp
    inprogress = false,
    whispopts = true,
    debuglevel = 3,
    promptrecur = true,
    mode=1,
  },
}

-------------------------------------------------------------------------------
--- INITIALIZE/ENABLE
-------------------------------------------------------------------------------
function flsys:OnInitialize()
  flsys.db = LibStub("AceDB-3.0"):New("FlamingLootSysDB", defaults)
  flsys.log = LibStub("AceDB-3.0"):New("flsysdebuglog")
  db = flsys.db.profile
end

function flsys:OnEnable()
  self:RegisterComm("flamingsysdata") -- for syncing data between mls
  self:RegisterChatCommand("fss", function() FLMain:Show() end)
  self:RegisterChatCommand("fs", "ProcessChatCmd")
  self:RegisterEvent("CHAT_MSG_ADDON")
  RegisterAddonMessagePrefix("flamingsys")
  flsys:GetConstants()
  flsys:InitializeFrames()
  if db.inprogress then -- this check needs to happen later
    flsys:InRaidCheck(true)
  end
  self:Print("Flaming loot masterlooter addon loaded.  Type /fs for a list of commands.")
  flsys:TrimDebug()
  flsys:CheckForGameUpdate()
end

function flsys:GetConstants()
  --load db constants into local memory
  raidstocount = db.raidstocount
  min_gp = db.min_gp
  ep_amt = db.ep_amt
  ep_recur = db.ep_recur
  standard_ilvl = db.standard_ilvl
  alts_data = db.alts_data
end

local function initialize_dropdown_all(self, level)
  local button = {}
  for k,v in pairs(db.sortlist) do
    button.text = v
    button.func = function(self)
      UIDropDownMenu_SetSelectedName(FLRosterDropDown, self:GetText());
    end
    UIDropDownMenu_AddButton(button, level)
  end
end

local function initialize_dropdown_perm(self, level)
  local button = {}
  for k,v in pairs(db.sortlist) do
    if db.players[v] and not db.players[v].temp then
      button.text = v
      button.func = function(self)
        UIDropDownMenu_SetSelectedName(FLRosterDropDown, self:GetText());
      end
      UIDropDownMenu_AddButton(button, level)
    end
  end
end

function flsys:InitializeFrames() --CALLS: MainButtons
  -- standings frame
  flsys:CheckRows()
  FLMain:SetScale(.8)
  FLMainPrev:GetNormalTexture():SetRotation(-math.pi/2)
  FLMainNext:GetNormalTexture():SetRotation(math.pi/2)
  local count = 0
  local texts = {" ", "EP", "GP", "PR"}
  local widths = {77,42,42,42}
  heading= CreateFrame("Button", "Heading" , FLMain)
  heading:SetSize(200,14)
  for i,w in ipairs(widths) do
    local c = FLMain:CreateFontString(nil, "ARTWORK", "GameFontNormal")
    c:SetParent(Heading)
    c:SetWidth(w)
    c:SetJustifyH("LEFT")
    c:SetPoint("TOPLEFT", FLMain, "TOPLEFT", 15+count, -42)
    c:SetText(texts[i])
    count = count+w
  end
  heading:Show()
  FLMainTitleText:SetText("         Standings")
  SetPortraitToTexture(FLMainPortrait, "Interface\\Icons\\Achievement_Arena_5v5_6")

  FLLogBox.ScrollBar:SetPoint("TOPRIGHT", FLLogContainer, "TOPRIGHT", -4, -17)
  FLLogItemAmount:SetNumeric(Enable)
  FLLogAdd:SetScale(.8)
  FLLogExport:SetScript("OnClick", function() flsys:ShowExportFrame("gear") end)

  -- export
  FLExport.TitleText:SetText("Copy and paste text to spreadsheet")
  FLExport.TitleText:SetJustifyH("LEFT")

  -- import/add
  FLImportData.TitleText:SetText("Import data")
  FLAddPlayer.TitleText:SetText("Add player")
  FLAddPlayer:SetScale(0.8)
  FLAddPlayerEP:SetNumeric(enable)
  FLAddPlayerGP:SetNumeric(enable)
  FLAddPlayerEP:SetNumber(0)
  FLAddPlayerGP:SetNumber(0)
  FLImportData:SetScale(.8)

  -- timers
  fra:SetScript("OnUpdate", function(self, elapsed) flsys:CheckTime(self, elapsed) end)
  fra:Hide()

  flsys:MainButtons()

  -- roster
  FLRosterFrame:SetScale(.8)
  FLRosterFrame.TitleText:SetText("Remove player")
  CreateFrame("Button", "FLRosterDropDown", FLRosterFrame, "UIDropDownMenuTemplate")
  FLRosterDropDown:SetPoint("LEFT")
  UIDropDownMenu_Initialize(FLRosterDropDown, initialize_dropdown_all)
end

function flsys:CheckForGameUpdate()
  -- ensure backwards compatibility of db.game_vers and db.standard_ilvl
  -- prompt to change standard ilvl if applicable
  new_user = (db.mode == 1 and #db.sortlist == 0) --hacky way to tell if this is a new user
  if not db.standard_ilvl then
    if new_user then
      db.standard_ilvl = DEFAULT_STANDARD_ILVL
    else
      db.standard_ilvl = 496 --assume 2.3.0 default bc since then it's been set
    end
  end

  local old_vers = db.game_vers or 0
  db.game_vers = string.match(select(1, GetBuildInfo()), "%d%.(%d)%.%d")

  if db.game_vers ~= old_vers and DEFAULT_STANDARD_ILVL > db.standard_ilvl then
    StaticPopup_Show("CHANGE_ILVL")
  end
end


local raid_check_timer = 0

function flsys:InRaidCheck(first_try)
  -- after login, if db.inprogress is true
  if IsInRaid() then
    StaticPopup_Show("RESUME_RAID")
  elseif first_try then
    delay_fra:SetScript("OnUpdate", function(self, elapsed) flsys:DelayRaidCheck(self, elapsed) end)
    delay_fra:Show()
  else
    flsys:EndRaid()
  end
end

function flsys:DelayRaidCheck(self, elapsed)
  raid_check_timer = raid_check_timer + elapsed
  if raid_check_timer > 2 then
    flsys:InRaidCheck(false)
    delay_fra:Hide()
  end
end

-------------------------------------------------------------------------------
--- START/END A RAID
-------------------------------------------------------------------------------
function flsys:StartRaid(raidname)
  local mlid = select(3, GetLootMethod())
  if db.inprogress then
    flsys:Print("You already have a raid in progress.")
  elseif not IsInRaid() then
    flsys:Print("You are not in a raid.")
  elseif not mlid then
    flsys:Print("You must be the masterlooter to start a raid.")
  elseif select(1, GetRaidRosterInfo(mlid)) ~= GetUnitName("player") then
    self:Print("You must be the masterlooter to start a raid.")
  else --prereqs met; start the raid
    self:RegisterEvent("GROUP_ROSTER_UPDATE")
    fld:Enable()
    db.inprogress = raidname or "Raid"
    flsys:MainButtons()
    table.insert(db.dates, 1, date("%m/%d %H:%M"))
    flsys:CreateRoster()
    flsys:Log(0, "start", "all", db.inprogress.." started", 0)  --FIX
    fld:StartRaid()
    flsys:SendVersion()
    if db.mode==1 then
      if db.promptrecur then
        StaticPopup_Show("START_RECUR")
      end
    elseif db.mode==2.1 then
      flsys:MassEP(4000, "single, GP")
    elseif db.mode==2.2 then
      flsys:MassEP(3, "single, count")
    end
  end
end

function flsys:ResumeRaid()
  flsys:UpdateRaid()
  if db.next_award>0 then
    flsys:Print("Recurring EP award resumed.")
    fra:Show()
  end
  flsys:RegisterEvent("GROUP_ROSTER_UPDATE")
  fld:Enable()
  fld:StartRaid()
end

function flsys:EndRaid()
  if db.inprogress then
    flsys:Log(0, "end", "all", "Raid ended") --FIX customize which raid
    self:Print("Raid ended.")
    db.dates[1] = db.dates[1].." - "..date("%m/%d %H:%M").." "..db.inprogress
    table.insert(db.dates, 1, "")
    db.inprogress=false
    for name, ntable in pairs(db.players) do
      if db.players[name].temp or db.players[name].alt then
        flsys:Debug(3, "Removing", name, "from system (alt or sub)")
        flsys:RemovePlayer(name)
      else
        db.players[name].inraid=nil
      end
    end
    flsys:EndRecur()
    self:UnregisterEvent("GROUP_ROSTER_UPDATE")
    self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
    self:UnregisterEvent("TRADE_ACCEPT_UPDATE")
    fld:Disable()
    flsys:RefreshSortList()
    if db.mode==1 then
      flsys:DecayRaid()
    end
  else
    self:Print("You do not have a raid in progress.")
  end
end

function flsys:DecayRaid()
  for name,_ in pairs(db.players) do
    db.players[name].ep_total = 0
    db.players[name].gp_total = 0
    local start = #db.players[name].ep_data
    if start<raidstocount then
      start = start+1
    end
    for i=start, 2, -1 do
      db.players[name].ep_data[i] = math.floor((db.players[name].ep_data[i-1] or 0)*(raidstocount-i+1)/(raidstocount-i+2))
      db.players[name].gp_data[i] = math.floor((db.players[name].gp_data[i-1] or 0)*(raidstocount-i+1)/(raidstocount-i+2))
      db.players[name].ep_total = db.players[name].ep_total+db.players[name].ep_data[i]
      db.players[name].gp_total = db.players[name].gp_total+db.players[name].gp_data[i]
    end
    db.players[name].ep_data[1] = 0
    db.players[name].gp_data[1] = 0
  end
  flsys:Print("EP and GP decayed.")
  flsys:Log(0, "decay", "all", "Decayed EP and GP")
  flsys:UpdatePR()
end

-------------------------------------------------------------------------------
--- RAID ROSTER
-------------------------------------------------------------------------------
function flsys:CreateRoster()
  numraidmembers = GetNumGroupMembers()
  flsys:Debug(3, "Creating roster, "..numraidmembers.." members")
  for i=1, numraidmembers do
    local name = select(1, GetRaidRosterInfo(i))
    if db.players[name] then
      db.players[name].inraid = true
    elseif db.alts_data[name] and db.players[db.alts_data[name]] then
      flsys:NewPlayer(name, true, false, true)
    else
      flsys:NewPlayer(name, true, true)
    end
  end
  flsys:RefreshSortList()
end

function flsys:GROUP_ROSTER_UPDATE()
  if not IsInRaid() then
    flsys:Debug(3, "GROUP_ROSTER_UPDATE - player no longer in raid")
    StaticPopup_Show("ENDRAID")
  elseif GetNumGroupMembers() ~=numraidmembers then
    flsys:Debug(3, "GROUP_ROSTER_UPDATE - # raid members changed")
    flsys:UpdateRaid()
    fld:CreateRoster()
  end
end

function flsys:UpdateRaid()
  for name, ntable in pairs(db.players) do
    db.players[name].inraid=nil
  end
  flsys:CreateRoster()
end

-------------------------------------------------------------------------------
--- MAIN FRAME DISPLAY
-------------------------------------------------------------------------------
function flsys:MainButtons() --portrait, buttons at bottom
  if not db.inprogress then
    FLMainButton1:SetText("Start")
    FLMainButton1:SetScript("OnClick", function() StaticPopup_Show("STARTRAID") end)
    FLMainButton2:SetText("Export")
    FLMainButton2:SetScript("OnClick", function() flsys:ShowExportFrame("standings") end)
  else
    FLMainButton1:SetText("End")
    FLMainButton1:SetScript("OnClick", function() StaticPopup_Show("ENDRAID") end)
    FLMainButton2:SetText("Mass EP")
    FLMainButton2:SetScript("OnClick", function() flsys:ShowMassEP() end)
  end
  flsys:UpdateMain()
end

function flsys:UpdateMain() --updates the player buttons - script & text
  for i, name in ipairs(db.sortlist) do
    local b = rowbuttons[i]
    if db.players[name].inraid then
      b.cells[1]:SetText("|cffCCFFFF"..name)
      b.cells[2]:SetText("|cffCCFFFF"..db.players[name].ep_total)
      b.cells[3]:SetText("|cffCCFFFF"..db.players[name].gp_total)
      b.cells[4]:SetText("|cffCCFFFF"..db.players[name].current_pr)
      if db.players[name].temp then
        b.cells[2]:SetText("|cffCCFFFF-")
        b.cells[4]:SetText("|cffCCFFFF-")
      end
    else
      b.cells[1]:SetText(name)
      b.cells[2]:SetText(db.players[name].ep_total)
      b.cells[3]:SetText(db.players[name].gp_total)
      b.cells[4]:SetText(db.players[name].current_pr)
    end
  end
  flsys:RosterPage(0)
end

function flsys:CheckRows()--creates & places row per sortlist if there isn't one already
  for i=1, #db.sortlist do
    if not rowbuttons[i] then
      local row = flsys:CreateTableRow({"-", "-", "-", "-"}, widths, FLMain)
      if i%13 ~= 0 then
        row:SetPoint("TOPLEFT", FLMain, "TOPLEFT", 18, -53-15*(i%13))
      else
        row:SetPoint("TOPLEFT", FLMain, "TOPLEFT", 18, -53-15*13)
      end
      rowbuttons[i] = row
    end
  end
end

function flsys:CreateTableRow(texttable)
  local widths = {77,42,42,42}
  local row = CreateFrame("Button", nil, FLMain)
  --row:SetHighlightTexture("Interface\\QuestFrame\\UI-QuestTitleHighlight", "ADD")
  row:SetSize(200,14)
  row:SetScript("OnClick", function() flsys:ShowLogFrame() end)
  row.cells = {}
  for i,w in ipairs(widths) do
    local c = row:CreateFontString(nil, "ARTWORK", "ChatFontSmall")
    c:SetJustifyH("LEFT")
    c:SetWidth(w)
    if #row.cells == 0 then
      c:SetPoint("LEFT", row, "LEFT", 0, 0)
    else
      c:SetPoint("LEFT", row.cells[#row.cells], "RIGHT", 0, 0)
    end
    c:SetText(texttable[i])
    table.insert(row.cells, c)
  end
  return row
end

function flsys:RosterPage(pg)
  rosterpage = rosterpage + pg
  for i, button in ipairs(rowbuttons) do
    if i<=#db.sortlist and i>rosterpage*13 and i<=(rosterpage+1)*13 then
      button:Show()
    else
      button:Hide()
    end
  end
  if rosterpage==0 then
    FLMainPrev:Hide()
  else
    FLMainPrev:Show()
  end
  if (rosterpage+1)*13>=#db.sortlist then
    FLMainNext:Hide()
  else
    FLMainNext:Show()
  end
end

-------------------------------------------------------------------------------
--- GP
-------------------------------------------------------------------------------
function flsys:OnLootAwarded(name, link, id, amount)
  flsys:Debug(3, "on loot awarded called: "..name..link..amount)
  if db.players[name] then
    flsys:AddGP(name, amount, link, id)
  end
end

function flsys:GetValue(level, itemid, equipLoc)
  -- function adapted from LibGearPoints-1.2 (EPGP)
  if db.mode~=2.2 then
    itemid=tonumber(itemid)
    if CUSTOM_ITEM_DATA[itemid] then
      level, equipLoc = unpack(CUSTOM_ITEM_DATA[itemid])
    end

    local slot_multiplier1 = EQUIPSLOT_MULTIPLIER_1[equipLoc]
    local slot_multiplier2 = EQUIPSLOT_MULTIPLIER_2[equipLoc]
    if not slot_multiplier1 then
      return 0
    end

    local gp_base = 1000*2^((level-standard_ilvl)/26)
    local high = math.floor(0.5 + gp_base * slot_multiplier1)
    local low = slot_multiplier2 and math.floor(0.5 + gp_base * slot_multiplier2) or nil
    flsys:Debug(3, "Returning "..high.." GP value")
    return high, low
  else
    return 1
  end
end

function flsys:AddGP(name, amount, itemname, id, raid_i)
  flsys:Debug(2, "Adding GP: "..name..amount..itemname)
  --scale
  local r_i = raid_i or 1
  amount = flsys:Round(amount*(raidstocount + 1 - r_i)/raidstocount, 0)
  --add to total & raid
  db.players[name].gp_data[r_i] = math.max(db.players[name].gp_data[r_i] + amount, 0)
  db.players[name].gp_total = math.max(db.players[name].gp_total + amount, 0)
  --don't add log if this is retroactive; let those funcs handle it
  if not raid_i then --only for auto (not log edits)
    flsys:Log(amount, "GP", name, itemname, id)
  end
  flsys:UpdatePR()
end

--administration
function flsys:ScaleGP(oldstd, newstd)
  local div = 2^((newstd-oldstd)/26)
  for name, nt in pairs(db.players) do
    nt.gp_total = flsys:Round(nt.gp_total/div)
    for i, g in ipairs(nt.gp_data) do
      nt.gp_data[i] = flsys:Round(g/div)
    end
  end
  flsys:Log(0, "decay", "all", "GP scaled: "..oldstd.." to "..newstd, 0)
  flsys:UpdatePR()
end

function flsys:ResetGP()
  for name, nt in pairs(db.players) do
    nt.gp_total=0
    for i, g in ipairs(nt.gp_data) do
      nt.gp_data[i] = 0
    end
  end
  flsys:Log(0, "RESET", "all", "Reset all GP")
  flsys:UpdatePR()
  self:Print("Reset all GP.")
end

-------------------------------------------------------------------------------
--- EP
-------------------------------------------------------------------------------
function flsys:AddEP(name, amount, reason, mass, raid_i)
  flsys:Debug(2, "Adding EP: "..name.." "..amount.."raid i "..(raid_i or "none"))
  -- if alt, add ep to main's record
  if db.players[name].alt then
    name = db.alts_data[name]
  end
  --scale based on raid index (assume current)
  local r_i = raid_i or 1
  amount = flsys:Round(amount*(raidstocount + 1 - r_i)/raidstocount, 0) --scale
  --update raid and total
  db.players[name].ep_data[r_i] = math.max(db.players[name].ep_data[r_i] + amount, 0)
  db.players[name].ep_total = math.max(db.players[name].ep_total + amount, 0)
  --update pr, add log entry
  if not mass then
    flsys:UpdatePR()
    if not raid_i then -- only for auto awards (not log edits)
      flsys:Log(amount, "EP", name, reason)
    end
  end
end

function flsys:MassEP(amount, reason)
  if db.inprogress then
    local namelist=""
    for name,_ in pairs(db.players) do
      if db.players[name].inraid then
        flsys:AddEP(name, amount, reason, true)
        namelist = namelist..name..", "
      end
    end
    flsys:Log(amount, "EP", string.sub(namelist, 1, -3), reason)
    flsys:UpdatePR()
    if db.mode==1 then
      flsys:Print(amount.."EP awarded to raid.")
    end
  else
    flsys:Print("There is no active raid.")
  end
end

function flsys:StartRecur(minutes) --should do this check before the popup
  if db.next_award>0 then
    flsys:Print("You already have a recurring award in progress.  /fs endrecur to end it.")
  elseif minutes>=0 and minutes<60 and db.inprogress then
    local start = GetTime()+minutes*60
    db.next_award=start
    flsys:Print("Recurring award starts in "..minutes.."min. "..ep_amt.." EP will be awarded every "..ep_recur.." min.")
    flsys:Debug(3, "Recurring award starts in "..minutes.."min. "..ep_amt.." EP will be awarded every "..ep_recur.." min.")
    fra:Show()
  elseif not db.sortlist[1] then
    flsys:Print("Recurring awards can only take place during an active raid.  /fs startraid to begin one.")
  else
    flsys:Print("Invalid entry.  Type /fs startrecur to bring up the window again.  Enter a number between 0 and 60.")
  end
end

function flsys:EndRecur()
  db.next_award=0
  flsys:Debug(3, "Ending recurring")
  fra:Hide()
end

local interval = 0
local currtime
function flsys:CheckTime(self, elapsed)
  --- adapted from the frame/func in epgp_recurring
  -- called by: fra.OnUpdate
  currtime = GetTime()
  interval = interval + elapsed
  if interval>30 then
    interval = 0
    if currtime > db.next_award then
      flsys:MassEP(ep_amt, db.inprogress)
      db.next_award = db.next_award + ep_recur*60
    end
  end
end

-------------------------------------------------------------------------------
--- PR & SORTING
-------------------------------------------------------------------------------
function flsys:UpdatePR()
  --- gets each players PR, sorts sortlist, sends standings 
  --BY: StartRaid, EndRaid, AddEP, AddGP, MassEP, RefreshSortList
  --CALLS: UpdateMain
  flsys:Debug(2, "SORT", "Updating PR, sorting")
  for name, i in pairs(db.players) do
    i.current_pr = flsys:GetPR(name)
  end
  table.sort(db.sortlist, function(name1, name2)
      return flsys:SortWrapper(name1, name2)
    end)
  if db.inprogress then
    flsys:SendStandings("raid")
  end
  flsys:MainButtons()
end

function flsys:GetPR(name)
  --- Returns a player's PR, 0 for player's without any
  --Called by: fld
  local pr
  if not ((db.mode==1 and db.players[name].temp) or db.players[name].alt) then
    if db.mode==2.2 then min_gp=.5 end
    local gp = math.max(db.players[name].gp_total, min_gp)
    local ep = db.players[name].ep_total
    pr = flsys:Round(ep/gp, 2)
  else
    pr = 0
  end
  return pr
end

function flsys:SortWrapper(name1, name2)
  if db.players[name1].inraid~=db.players[name2].inraid then
    return db.players[name1].inraid
  elseif db.players[name1].temp~=db.players[name2].temp then
    return db.players[name2].temp
  else
    return flsys:SortByPR(name1, name2)
  end
end

function flsys:SortByPR(name1, name2)
  if db.players[name1].current_pr ~= db.players[name2].current_pr then
    return db.players[name1].current_pr > db.players[name2].current_pr
  else
    return db.players[name1].gp_total < db.players[name2].gp_total
  end
end

function flsys:RefreshSortList()
  --- Wipes & recreates sortlist containing all players
  -- called by: CreateRoster --TODO
  -- calls: CheckRows, UpdatePR
  wipe(db.sortlist)
  for name,t in pairs(db.players) do
    if t then table.insert(db.sortlist, name) end
  end
  flsys:CheckRows()
  flsys:UpdatePR()
end

-------------------------------------------------------------------------------
--- SEND TO RAIDERS
-------------------------------------------------------------------------------
function flsys:SendStandings(sender, uptodate)
  if uptodate then
    SendAddonMessage("flamingloot", "uptodate:", "WHISPER", sender)
  else
    sstr =""
    for i, name in pairs(db.sortlist) do
      sstr = sstr..name..": "
      if db.mode~=2.2 then
        sstr = sstr..db.players[name].current_pr.."\n"
      else
        sstr = sstr..db.players[name].gp_total.."\n"
      end
    end
    if sender=="raid" then
      SendAddonMessage("flamingloot", "standings:"..sstr, "RAID")
      db.lastbroadcast = time()
    else
      SendAddonMessage("flamingloot", "standings:"..sstr, "WHISPER", sender)
    end
  end
end

function flsys:SendVersion()
  local vers = GetAddOnMetadata("flamingloot", "Version")
  local one, two = string.match(vers, "(%d).*(%d)")
  if one and two then
    SendAddonMessage("flamingloot", "version:"..one..two, "RAID")
    flsys:Debug(3, "VERSION", "Sent check to raid: version:"..one..two)
  end
end

function flsys:CHAT_MSG_ADDON(event, prefix, message, channel, sender)
  if prefix=="flamingsys" then
    local command, value = string.match(message, "(%a+):*(.*)")
    if command=="datareply" then
      self:Print(sender.." "..value)
    elseif command=="standings" then
      flsys:SendStandings(sender, (tonumber(value))<db.lastbroadcast)
    end
  end
end

function flsys:OnCommReceived(prefix, message, channel, sender)
  if prefix=="flamingsysdata" then
    local command, value = string.match(message, "(%a+):*(.*)")
    if command=="data" then
      local dialog = StaticPopup_Show("RECEIVE_DATA", sender)
      if dialog then
        dialog.text = value
        dialog.player = sender
      end
    end
  end
end

-------------------------------------------------------------------------------
--- CHAT COMMANDS
-------------------------------------------------------------------------------
function flsys:ProcessChatCmd(input)
  if not input or input:trim() == "" then
    self:Print("Addon commands:\n|cffffff78/fss|r = View the main panel.\n|cffffff78/fs opt|r = View the options window. Also contains a help panel.\n|cffffff78/fs gp [ilvl]|r = View a printout of GP values for an ilvl.\n|cffffff78/fs decay|r = Decay ep and gp (default setting will prompt after raids so you don't need to use this).\n|cffffff78/loot|r = (during a raid only) View a cache of the most recent mob's drops and all the rolls made on them.  Only available after the rolling period ends and only for the most recent mob.")
  elseif input=="opt" then
    if not flopt:IsEnabled() then
      fld:Enable()
      flopt:Enable()
    end
    InterfaceOptionsFrame_OpenToCategory("flaming loot ml")
  elseif input=="decay" then
    if not db.inprogress then
      StaticPopup_Show("DECAY")
    else
      StaticPopup_Show("NOTDURINGRAID")
    end
  else
    local ilvl = tonumber(string.match(input, "gp (%d+)"))
    if ilvl then
      if (-100 < ilvl - standard_ilvl) and (ilvl - standard_ilvl < 100) then --reasonable-ish number
        flsys:PrintGPValues(ilvl)
      else
        self:Print("Not a valid item level. Try again.")
      end
    else
      self:Print("Not a valid command. Type /fs to view a list of commands")
    end
  end
end

function flsys:PrintGPValues(ilvl)
  local mult = flsys:Round((2^((ilvl-standard_ilvl)/26)), 3)
  self:Print(ilvl.." values:\n2H weapons, hunter ranged: "..(2000*mult).."\nCaster/healer main hand, tank shield: "..(1500*mult).."\nDual wield weapon, chest, helm, legs, trinket: "..(1000*mult).."\nShoulder, belt, boots, gloves: "..(750*mult).."\nOff hand, tank mainhand, neck, ring, cloak, bracers: "..(500*mult))
end

-------------------------------------------------------------------------------
--- LOG DISPLAY
-------------------------------------------------------------------------------
function flsys:ShowLogFrame()
  if #log_rows == 0 then --create frames if needed
    local fs
    for i = 1, 13 do
      log_rows[i] = CreateFrame("Button", "LogRow"..i, FLLogContainer, "LogRowTemplate")
      log_rows[i]:SetPoint("TOPLEFT", FLLogContainer, "TOPLEFT", 5, -18*i+15)
      log_rows[i]:SetMotionScriptsWhileDisabled(true)
      fs = log_rows[i]:GetFontString()
      fs:SetJustifyH("LEFT")
    end
    --start it scrolled down to newest content
    flsys:Debug(3, "LOG DISPLAY", "created rows, calling on vert scroll")
    local starting_os = math.max(#db.epgplog - 13, 0)
    flsys:UpdateLogFrame(starting_os)
    FLLog:Show()
    FLLogBoxScrollBar:SetValue(starting_os*18)
  elseif not FLLog:IsShown() then
    flsys:UpdateLogFrame()
    FLLog:Show()
  end
end

function flsys:UpdateLogFrame(offset)
  --- given the position of the scroll, sets each log row's text and onclick
  -- @offset: given in number of rows. calculated by OnVerticalScroll
  -- calls: FauxScrollFrame_Update
  -- called by: ShowLogFrame, OnVerticalScroll

  local num_logs = #db.epgplog
  --print(num_logs)
  local start_raid_i = 1 --for decay calculations. 1 = current
  if num_logs > 0 then
    for i = 1, num_logs - (offset or 0) do
      if db.epgplog[i][3]=="decay" then
        start_raid_i = start_raid_i + 1
      end
    end
  end
  for i = 1, 13 do
    --print(num_logs.." "..(i-1).." "..(offset or 0))
    local log_i = num_logs - (i-1) - (offset or 0)
    local raid_i = start_raid_i -- so on script don't all call the last version
    --print("log i "..log_i..", raid_i "..raid_i)
    if log_i > 0 then
      local logitem = db.epgplog[log_i]
      log_rows[i]:SetText(flsys:LogItemtoString(logitem))
      fs = log_rows[i]:GetFontString()
      if fs:IsTruncated() then
        log_rows[i]:SetScript("OnEnter", function(self)
          GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
          GameTooltip:SetText(self:GetText())
          GameTooltip:Show()
        end)
        log_rows[i]:SetScript("OnLeave", function() GameTooltip:Hide() end)
      else
        log_rows[i]:SetScript("OnEnter", nil)
        log_rows[i]:SetScript("OnLeave", nil)
      end
      if logitem[3]=="GP" or (logitem[3]=="EP" and not string.find(logitem[4], ',')) then
        -- can edit/delete
        --flsys:Print(i.."gp or single ep, setting texture")
        log_rows[i]:Enable()
        log_rows[i]:SetNormalTexture("Interface\\Buttons\\UI-GuildButton-PublicNote-Up")
        log_rows[i]:SetPushedTexture("Interface\\Buttons\\UI-GuildButton-PublicNote-Up")
        log_rows[i]:SetScript("OnClick", function() flsys:ShowEditLogItem(log_i, raid_i); end)
      elseif logitem[3]=="start" then
        -- can add
        log_rows[i]:Enable()
        log_rows[i]:SetNormalTexture("Interface\\Buttons\\UI-AttributeButton-Encourage-Up")
        log_rows[i]:SetPushedTexture("Interface\\Buttons\\UI-AttributeButton-Encourage-Down")
        log_rows[i]:SetScript("OnClick", function() flsys:ShowAddType(log_i, raid_i); end)
      else
        log_rows[i]:Disable()
        if logitem[3] == "decay" then
          start_raid_i = start_raid_i - 1
        end
      end
      if not log_rows[i]:IsShown() then
        log_rows[i]:Show()
      end
    else
      -- extra row doesn't correspond to any log entry; hide it
      log_rows[i]:Hide()
    end
  end
  FauxScrollFrame_Update(FLLogBox, num_logs, 13, 18)
end

function flsys:OnVerticalScroll(self, value, itemHeight)
	FLLogBoxScrollBar:SetValue(value)
  offset = floor((value / itemHeight) + 0.5);
	self.offset = offset
  flsys:UpdateLogFrame(offset)
end

function flsys:LogItemtoString(logitem)
  local displaystring = "  "
  local logdate, amount, logtype, players, reason, num = unpack(logitem)
  if logtype=="start" or logtype=="import" or logtype=="RESET" then
    displaystring = "|cffCCFFFF"..date("%m/%d/%y %I:%M%p", logdate).." "..reason.." "
    displaystring = displaystring..string.rep("-", 62-#displaystring).."|r"
  elseif logtype=="end" or logtype=="decay" then
    displaystring = displaystring..date("%I:%M%p", logdate).."  "..reason
  else
    displaystring = displaystring..date("%I:%M%p", logdate).."  "
    if logtype=="EP" then
      if amount>0 then amount= "+"..amount end
      displaystring = displaystring..amount.." EP "..players.." ("..reason..")"
    elseif logtype=="GP" then
      displaystring = displaystring..reason.." "..players.." ("..amount.." GP)"
    end
  end
  return displaystring
end

function flsys:ShowAddType(log_i, raid_i)
  FLLogAddEP:SetScript("OnClick", function() flsys:ShowAddToLog(log_i, raid_i, "EP") end)
  FLLogAddGP:SetScript("OnClick", function() flsys:ShowAddToLog(log_i, raid_i, "GP") end)
  --FLLogAddMassEP:SetScript("OnClick", function() flsys:ShowAddToLog(log_i, raid_i, "EPmass") end)

  FLLogAdd:Show()
end

local function setup_dropdown(raid_i)
  FLRosterDropDown:ClearAllPoints()
  FLRosterDropDown:SetParent(FLLogItem)
  FLRosterDropDown:SetPoint("BOTTOMLEFT", FLLogItem, "BOTTOMLEFT", 70, 30)

  if raid_i == 1 then
    UIDropDownMenu_Initialize(FLRosterDropDown, initialize_dropdown_all)
  else
    UIDropDownMenu_Initialize(FLRosterDropDown, initialize_dropdown_perm)
  end
end

function flsys:ShowAddToLog(log_i, raid_i, award_type)
  --- Shows frame for adding an award retroactively
  -- @log_i: the index of the epgplog item where the raid started
  -- @raid_i: raid index, used to scale and place the award appropriately
  -- @award_type: ep or gp

  FLLogAdd:Hide()
  --flsys:Print('Retro add '..award_type..", raid i "..raid_i)
  local start = db.epgplog[log_i]
  FLLogItem.TitleText:SetText("Adding award to "..date("%m/%d/%y", start[1])
                               .." "..string.sub(start[5], 1, -8))
  FLLogItemAmount:SetNumber(0)
  FLLogItemType:SetText(award_type)
  FLLogItemReason:SetText("")
  FLLogItem:SetHeight(105)
  FLLogItemNote:SetText("")

  setup_dropdown(raid_i)

  FLLogItemDelete:Hide()
  FLLogItemUpdate:SetScript("OnClick", function() flsys:AlterLog(log_i, raid_i, "add") end)
  FLLogItem:Show()
end

function flsys:ShowEditLogItem(log_i, raid_i)
  --- fills in and shows the frame for editing/deleting a log entry

  logitem = db.epgplog[log_i]
  --flsys:Print("raid index "..raid_i..", log: "..log_i..flsys:LogItemtoString(logitem))
  FLLogItem.TitleText:SetText("Editing award "..date("%m/%d/%y %H:%M", logitem[1]))
  FLLogItemAmount:SetNumber(logitem[2])
  FLLogItemType:SetText(logitem[3])
  FLLogItemReason:SetText(logitem[5])
  FLLogItem:SetAttribute("raid", raid_i) --TODO TEST
  FLLogItem:SetAttribute("log", log_i)

  setup_dropdown(raid_i)
  if db.players[logitem[4]] and (raid_i == 1 or not db.players[logitem[4]].temp) then
    FLLogItem:SetHeight(105)
    FLLogItemNote:SetText("")
    UIDropDownMenu_SetSelectedName(FLRosterDropDown, logitem[4])
  else
    FLLogItem:SetHeight(145)
    if alts_data[logitem[4]] and logtype[3] == "GP" then
      -- EP awards for alts are stored under main's name, but GP are not
      FLLogItemNote:SetText("Note: This was a GP award to an alt, which has no impact on ongoing PR.")
    else
      FLLogItemNote:SetText("Note: This player is no longer in the database.  Editing the amount or reason won't impact anyone's PR, but you can reassign the award to another player.")
    end
  end

  FLLogItemDelete:SetScript("OnClick", function()
    local dialog = StaticPopup_Show("DELETE_LOG_ITEM")
    if dialog then
      dialog.log_i = log_i
      dialog.raid_i = raid_i
    end
  end)
  FLLogItemDelete:Show()
  FLLogItemUpdate:SetScript("OnClick", function() flsys:AlterLog(log_i, raid_i, "edit") end)
  FLLogItem:Show()
end

-------------------------------------------------------------------------------
--- ALTER LOG
-------------------------------------------------------------------------------
function flsys:Log(amount, logtype, player, reason, num)
  --- logs an event. updates log frames as needed.

  local timestamp = time()
  table.insert(db.epgplog, 1, {timestamp, amount, logtype, player, reason, num})
  -- update log frame if needed
  if FLLog:IsShown() then
    flsys:UpdateLogFrame()
  end
  -- update log edit/add if needed
  if FLLogItem:IsShown() then
    local raid_i = FLLogItem:GetAttribute("raid")
    local log_i = FLLogItem:GetAttribute("log") + 1
    if logtype=="decay" then
      raid_i = raid_i + 1
    end
    flsys:ShowEditLogItem(log_i, raid_i)
  end
end

function flsys:AlterLog(log_i, raid_i, alter_type)
  --- takes form input and updates the log, ep & gp accordingly
  --- uses Undo, AddEP and AddGP to handle epgp, modifies the log itself
  -- @alter_type: "edit" or "add"

  -- get original (used for editing)
  old_item = db.epgplog[log_i]
  -- get input
  local amount = FLLogItemAmount:GetNumber()
  local log_type = FLLogItemType:GetText()
  local player = UIDropDownMenu_GetSelectedName(FLRosterDropDown)
  local reason = FLLogItemReason:GetText()
  --validate fields are filled out
  if not (amount and player and reason) then
    flsys:Print("Please fill out all the fields.")
    return
  end
  --if edit, check whether undo is needed
  if alter_type == "edit" then
    if amount == old_item[2] and player == old_item[4] then
      -- don't need to touch playerdb, just update reason in log
      db.epgplog[log_i][5] = reason
      flsys:Print("No changes made to ep or gp.  Log modified")
      return
    else
      flsys:Undo(log_i, raid_i, false)
    end
  end
  -- add to new player
  if log_type == "EP" then
    flsys:AddEP(player, amount, reason, false, raid_i)
  elseif log_type == "GP" then
    flsys:AddGP(player, amount, reason, 0, raid_i)
  end
  if alter_type == "edit" then -- update log entry
    db.epgplog[log_i][2] = amount
    db.epgplog[log_i][4] = player
    db.epgplog[log_i][5] = reason
    flsys:Print("Change saved.")
  else -- add log entry
    -- insert before log_i so it is more recent than the raid start
    if raid_i == 1 then --chrono order for current raid
      log_i = 1
    end
    table.insert(db.epgplog, log_i, {time(), amount, log_type, player, reason})
    flsys:Print("Award added.")
  end
  FLLogItem:Hide()
  flsys:UpdateLogFrame()
end

function flsys:Undo(log_i, raid_i, remove)
  --- Undo an existing EP or GP award.
  -- @remove indicates whether to delete the log entry & update everything
  --    (true when deleting, false when part of an edit)
  --print("undo - raid i "..raid_i..", log i "..log_i)
  local _, amount, logtype, player, _, _ = unpack(db.epgplog[log_i])
  amount = flsys:Round(amount*(raidstocount + 1 - raid_i)/raidstocount, 0) --scale
  if logtype=="EP" then
    for name in string.gmatch(player, "([^, ]+)") do
      if db.players[player] then
        db.players[name].ep_total = db.players[name].ep_total - amount
        db.players[name].ep_data[raid_i] = db.players[name].ep_data[raid_i] - amount
      else
        flsys:Print("Could not undo EP award for "..player.."; they were not found in the database")
      end
    end
  elseif logtype=="GP" then
    if db.players[player] then
      db.players[player].gp_total = db.players[player].gp_total - amount
      db.players[player].gp_data[raid_i] = db.players[player].gp_data[raid_i] - amount
    else
      flsys:Print("Could not undo GP for "..player.."; they are no longer in the database.")
      return
    end
  end
  if remove then
    table.remove(db.epgplog, log_i)
    flsys:UpdateLogFrame() --refresh it
    flsys:UpdatePR() --refresh standings
  end
end

function flsys:ClearLog(logtype)
  FLLog:Hide()
  if not logtype or logtype=="epgp" then
    wipe(db.epgplog)
    flsys:Print("EPGP log cleared")
  end
  -- don't need to call UpdateLog; it will get called when log is shown
end

-------------------------------------------------------------------------------
--- PLAYER DATABASE
-------------------------------------------------------------------------------
function flsys:AddPlayerEntered(done)
  --- Handles frame parsing and updating, checks for error cases,
  --- then passes on to NewPlayer for creation
  --CALLS: ~NewPlayer, ~InitializeFrames
  local name = flsys:Cap(FLAddPlayerPlayer:GetText())
  local ep = FLAddPlayerEP:GetNumber()
  local gp = FLAddPlayerGP:GetNumber()
  if name=="" then
    self:Print("Please enter a player name.")
  elseif db.players[name] and not db.players[name].temp then
    self:Print(name.." is already in the system.  If you want to reset their data, remove them and then add.")
  else
    flsys:Debug(3, "Adding player: "..name..", EP: "..ep..", GP: "..gp)
    flsys:NewPlayer(name, nil, nil, nil, ep, gp)
    FLAddPlayerPlayer:SetText("")
    FLAddPlayerEP:SetNumber(0)
    FLAddPlayerGP:SetNumber(0)
    if done then
      FLAddPlayer:Hide()
      flsys:InitializeFrames()
    end
  end
end

function flsys:NewPlayer(name, inraid, temp, alt, ep, gp)
  --- Adds a player to the database
  -- calls: RefreshSortList, ~UpdateRaid
  db.players[name] = { --applies to all
    inraid = inraid,
    temp = temp or false,
    alt = alt,
    balance = 0,
    ep_data = {[1] = 0},
    gp_data = {[1] = 0},
  }
  if ep and gp then
    db.players[name].ep_total = flsys:Round(ep*(raidstocount-1)/(raidstocount))
    db.players[name].gp_total = flsys:Round(gp*(raidstocount-1)/(raidstocount))
    db.players[name].gp_data[2] = db.players[name].gp_total
    db.players[name].ep_data[2] = db.players[name].ep_total
  else
    db.players[name].ep_total = 0
    db.players[name].gp_total = 0
  end
  if db.inprogress then
    flsys:UpdateRaid()
  end
  flsys:RefreshSortList()
end

function flsys:RemovePlayer(name)
  FLRosterFrame:Hide()
  if db.inprogress then
    self:Print("Cannot remove players during a raid.")
  else
    flsys:Debug(3, "Removing "..name.." from system.")
    db.players[name]=nil
    if db.mode==1 then self:Print(name.." removed from database.") end
    flsys:RefreshSortList()
  end
end

function flsys:AddAlt(info, input)
  local aremove, alt, main = string.match(input, "(-*)([^%d%s%p]+)=*([^%d%s%p]*)")
  if alt then alt = flsys:Cap(alt) end
  if main then main = flsys:Cap(main) end
  if aremove and db.alts_data[alt] then
    db.alts_data[alt] = nil
    flsys:Print(alt.." removed from alts list.")
  elseif alt and main then
    if not db.players[main] then
      self:Print(main.." is not in the database.  Add them before adding their alts.")
    else
      db.alts_data[alt] = main
      flsys:Print(alt.." added to alts list.")
    end
  else
    flsys:Print("Invalid entry. Make sure to format 'altname=mainname' without spaces")
  end
  alts_data = db.alts_data
end

-------------------------------------------------------------------------------
--- MAJOR SETTINGS CHANGES
-------------------------------------------------------------------------------
function flsys:ModeChanged(mode)
  if mode==1 then
    if not db.players[GetUnitName("player")] then
      flsys:NewPlayer(GetUnitName("player"))
    end
  else
    flsys:ResetData(true)
  end
end

function flsys:ResetData(removeplayers)
  wipe(db.dates)
  wipe(db.epgplog)
  if removeplayers then
    wipe(db.players)
    wipe(db.alts_data)
    wipe(alts_data)
    self:Print("Removed all players and their data from the system.")
    flsys:Log(0, "RESET", "all", "Reset & removed all players")
  else
    for name, ntable in pairs(db.players) do
      flsys:NewPlayer(name)
    end
    self:Print("Reset data for all players.")
    flsys:Log(0, "RESET", "all", "Reset all EP and GP")
  end
  flsys:RefreshSortList()
end

-------------------------------------------------------------------------------
--- EXPORT
-------------------------------------------------------------------------------
function flsys:ShowExportFrame(exporttype)
  -- 1 timestamp 2 amount 3 type 4 player(s) 5 reason/item 6 itemid
  --spreadsheet order: 1 4 5 2 3 (6)
  local exporttext = ""
  if exporttype=="gear" then
    for i, entry in ipairs(db.epgplog) do
      exporttext = exporttext..date("%m/%d/%y %H:%M", entry[1]).."\t"..entry[4].."\t"..entry[5].."\t"..entry[2].."\t"..entry[3]
      if entry[6] then
        exporttext = exporttext.."\t"..entry[6]
      end
      exporttext = exporttext.."\n"
    end
    FLExportDone:SetScript("OnClick", function() FLExport:Hide(); flsys:ConfirmExportTime("epgp") end)
  elseif exporttype=="standings" then
    local ddate = date("%m/%d/%y %H:%M", time())
    for i, name in ipairs(db.sortlist) do
      exporttext = exporttext..name.."\t"..db.players[name].ep_total.."\t"..db.players[name].gp_total.."\t"..db.players[name].current_pr.."\t"..ddate
      if i==1 then
        exporttext = exporttext.."\t"..flsys:GetImportText()
      end
      exporttext = exporttext.."\n"
    end
    FLExportDone:SetScript("OnClick", function() FLExport:Hide(); end)
  end
  if exporttext~="" then
    FLExportText:SetText(exporttext)
    FLExport:Show()
  else
    self:Print("Nothing to export.")
  end
end

function flsys:GetImportText() --not during combat?
  --- Provides a string that could be used to import all player epgp data
  local importstr =""
  for name, nt in pairs(db.players) do
    importstr = importstr.."@"..name.."=gp:"
    for i, amt in ipairs(db.players[name].gp_data) do
      importstr = importstr..amt..","
    end
    importstr = importstr.."ep:"
    for i, amt in ipairs(db.players[name].ep_data) do
      importstr = importstr..amt..","
    end
  end
  return importstr
end

function flsys:ConfirmExportTime(exporttype)
  --- Asks the player to confirm export complete so we can trim log
  dialog = StaticPopup_Show("EXPORT_COMPLETE", exporttype)
  if dialog then
    dialog.etype = exporttype
  end
end

-------------------------------------------------------------------------------
--- IMPORT
-------------------------------------------------------------------------------
function flsys:Import()
  InterfaceOptionsFrame:Hide()
  FLImportDataData:SetText("")
  FLImportData:Show()
end

function flsys:ImportEntered(imptext, fromplayer)
  if db.inprogress then
    self:Print("You cannot import or receive data while you have a raid in progress.")
    return
  end
  if imptext and imptext~="" then
    local printmsg, sendplayer, sendplayererror
    for psection in string.gmatch(imptext, "@([^@]+)") do
      local player = string.match(psection, "^([^=]+)")
      local eppos = string.find (psection, "ep:")
      local balpos = string.find (psection, "bal:") or -1
      if not player or not eppos or not string.find(psection, "gp:") then
        printmsg = "Import text not formatted correctly. Make sure you paste the entire text exactly as it is in the spreadsheet.  If you're seeing this error after receiving data from another player, please go to curse.com/addons/wow/flaming-ml and report a bug."
        sendplayererror = "could not import your data correctly.  This is probably an addon bug.  Please go to curse.com/addons/wow/flaming-ml and report it."
      else
        FLImportData:Hide()
        flsys:NewPlayer(player)
        local currpart = string.sub(psection, 1, eppos)
        local pos = 1
        for amt in string.gmatch(currpart, "(%d+)") do --filling in gp data, updating total
          amt = tonumber(amt)
          db.players[player].gp_data[pos] = amt
          db.players[player].gp_total = db.players[player].gp_total +amt
          pos = pos+1
        end
        currpart = string.sub(psection, eppos, balpos)
        pos = 1
        for amt in string.gmatch(currpart, "(%d+)") do --filling in epdata, updating total
          amt = tonumber(amt)
          db.players[player].ep_data[pos] = amt
          db.players[player].ep_total = db.players[player].ep_total +amt
          pos = pos+1
        end
        local bal = string.match(psection, "bal:(%-?%d+)")
        if bal then
          db.players[player].balance = tonumber(bal)
        end
        flsys:Log(0, "import", player, "Imported data for "..player)
        self:Print("Imported data for "..player..".")
        flsys:UpdatePR()
        sendplayer="has imported your data!"
      end
    end
    self:Print(printmsg)
    if fromplayer then
      if sendplayererror then
        SendAddonMessage("flamingsys", "datareply:"..sendplayererror, "WHISPER", fromplayer)
      elseif sendplayer then
        SendAddonMessage("flamingsys", "datareply:"..sendplayer, "WHISPER", fromplayer)
      end
    end
  else
    self:Print("You entered a blank import.  Copy the text from your data backup spreadsheet.  If you're seeing this error after trying to receive data from another player, confirm that they have data stored - if so, please go to curse.com/addons/wow/flaming-ml and report a bug.")
    if fromplayer then
      SendAddonMessage("flamingsys", "datareply: did not receive your data - either you have no data stored or there was an addon bug.  Please go to curse.com/addons/wow/flaming-ml and report it.", "WHISPER", fromplayer)
    end
  end
end

function flsys:SendData(name)
  self:SendCommMessage("flamingsysdata", "data:"..flsys:GetImportText(), "WHISPER", name)
  self:Print("Sending data to "..name..".  If they accept or decline the data you'll see a printout.  No further message means they don't have the addon turned on or they're using an outdated version.")
end

-------------------------------------------------------------------------------
--- MISC UTILITIES
-------------------------------------------------------------------------------
function flsys:Round(num, dp)
  local mult = 10^(dp or 0)
  return math.floor(num * mult + 0.5) / mult
end

function flsys:TextTooltip(owner, text)
  GameTooltip:SetOwner(owner, "ANCHOR_RIGHT")
  GameTooltip:SetText(text)
  GameTooltip:Show()
end

function flsys:Print(text)
  if text then
    DEFAULT_CHAT_FRAME:AddMessage("|cff6600FFflsys:|r "..tostring(text))
  end
end

function flsys:Cap(str)
  local capstr = ""
  for word in string.gmatch (str, "[^%s]+") do
    if word~="of" and word~="the" and word~="and" and word~="a" and word~="or" then
      word = (word:gsub("^%l", string.upper))
    end
    capstr = capstr..word.." "
  end
  return string.sub(capstr, 1, -2)
end

function flsys:Debug(lvl, logtype, ...)
  if lvl<=db.debuglevel then
    table.insert(flsys.log.profile, date("%m/%d %H:%M:%S").."---"..logtype.."---"..(table.concat({...}, " ")))
  end
end

function flsys:TrimDebug()
  for i = 1001, #flsys.log.profile do
    table.remove(flsys.log.profile, 1)
  end
end

-------------------------------------------------------------------------------
--- POPUP CALLS TODO are these necessary, could just do _show directly?
-------------------------------------------------------------------------------
function flsys:AsktoSetupRecur()
  if db.next_award==0 then
    StaticPopup_Show("START_RECUR")
  end
end

function flsys:CheckEndRecur()
  StaticPopup_Show("ENDRECUR")
end

function flsys:DecayPressed()
  StaticPopup_Show("DECAY")
end

function flsys:ResetDataPressed() --inprogress check done
  StaticPopup_Show("DATARESET")
end

function flsys:ResetLogPressed()
  StaticPopup_Show("LOGRESET")
end

function flsys:EndPressed()
  StaticPopup_Show("ENDRAID")
end

function flsys:RemovePlayerPressed(player)
  if player then
    local dialog = StaticPopup_Show("REMOVEPLAYER", player)
    if dialog then
      dialog.player = player
      dialog.enterClicksFirstButton = true
    end
  end
end

function flsys:ResetGPPressed()
  StaticPopup_Show("RESET_GP")
end

-------------------------------------------------------------------------------
--- POPUP DIALOGS
-------------------------------------------------------------------------------
StaticPopupDialogs["MASSEP"] = {
  text = "Mass EP",
  button1 = "Single",
  button2 = CANCEL,
  button3 = "Recurring",
  hideOnEscape = 1,
  hasEditBox = 1,
  OnAccept = function(self, data)
    amt = tonumber(self.editBox:GetText())
    if amt and amt>0 then
      flsys:MassEP(amt, db.inprogress)
    else
      self.text = "Please enter a number amount of EP to award",
      self:Show()
    end
  end,
  OnAlt = function(self, data)
    amt = tonumber(self.editBox:GetText())
    if amt and amt>0 then
      flsys:StartRecur(0)
    else
      self.text = "Please enter a number amount of EP to award",
      self:Show()
    end
  end
}

StaticPopupDialogs["DATARESET"] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = "Also remove all players from the database?",
  button1 = "Yes",
  button2 = "No",
  button3 = "Cancel",
  timeout = 0,
  hideOnEscape = 1,
  whileDead = 1,
  OnAccept = function(self) flsys:ResetData(true) end,
  OnCancel = function(self) flsys:ResetData(false) end,
}

StaticPopupDialogs["LOGRESET"] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = "Delete all log info?",
  button1 = ACCEPT,
  button2 = CANCEL,
  timeout = 0,
  hideOnEscape = 1,
  whileDead = 1,
  OnAccept = function(self) flsys:ClearLog() end,
}

StaticPopupDialogs["RESET_GP"] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = "Reset all players' GP to 0?  This cannot be undone.",
  button1 = ACCEPT,
  button2 = CANCEL,
  timeout = 0,
  hideOnEscape = 1,
  whileDead = 1,
  OnAccept = function(self) flsys:ResetGP() end,
}

StaticPopupDialogs["STARTRAID"] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = "Name of raid:",
  button1 = ACCEPT,
  button2 = CANCEL,
  timeout = 0,
  whileDead = 1,
  hideOnEscape = 1,
  hasEditBox = 1,
  maxLetters = 25,
  OnAccept = function(self)
    local text = self.editBox:GetText()
    flsys:StartRaid(flsys:Cap(text))
  end
}

StaticPopupDialogs["ENDRAID"] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = "End the raid?",
  button1 = ACCEPT,
  button2 = CANCEL,
  hideOnEscape = 1,
  timeout = 0,
  whileDead = 1,
  OnAccept = function(self) flsys:EndRaid() end,
  OnCancel = function()
    if not fld:IsEnabled() then
      flsys:ResumeRaid()
    end
  end,
}

StaticPopupDialogs["REMOVEPLAYER"] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = "Remove %s from the system? All of their data will be lost.",
  button1 = "Yes",
  hideOnEscape = 1,
  button2 = "No",
  timeout = 0,
  whileDead = 1,
  OnAccept = function(self)
    flsys:RemovePlayer(self.player)
  end,
}

StaticPopupDialogs["START_RECUR"] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = "Number of minutes until first award should go out:",
  button1 = ACCEPT,
  button2 = CANCEL,
  timeout = 0,
  whileDead = 1,
  enterClicksFirstButton = true,
  hasEditBox = 1,
  maxLetters = 2,
  OnShow = function(self, data)
    self.editBox:SetText("0")
  end,
  OnAccept = function(self, data)
    local minutes = tonumber(self.editBox:GetText())
    flsys:StartRecur(minutes)
  end
}

StaticPopupDialogs["ENDRECUR"] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = "End recurring EP award?",
  button1 = "Yes",
  button2 = "No",
  timeout = 0,
  enterClicksFirstButton = true,
  whileDead = 1,
  hideOnEscape = 1,
  OnAccept = function(self)
    flsys:EndRecur() end,
}

StaticPopupDialogs["RESUME_RAID"] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = "Resume your raid?",
  button1 = "Yes",
  button2 = "No",
  timeout = 0,
  whileDead = 1,
  OnAccept = function(self) flsys:ResumeRaid() end,
  OnCancel = function(self) flsys:EndRaid() end,
}

StaticPopupDialogs["NOTDURINGRAID"] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = "You cannot do that while a raid is in progress.",
  button1 = "Ok",
  timeout = 0,
  hideOnEscape = 1,
  enterClicksFirstButton = true,
  whileDead = 1,
}

StaticPopupDialogs["CONFIRM_RAIDSTOCOUNT"] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = "Change number of raids to count?",
  button1 = "Yes",
  button2 = CANCEL,
  timeout = 0,
  enterClicksFirstButton = true,
  hideOnEscape = 1,
  whileDead = 1,
  OnAccept = function(self)
    flsys.db.profile.raidstocount = self.input
    flsys:GetConstants()
    end,
}

StaticPopupDialogs["EXPORT_COMPLETE"] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = "Did you successfully back up your data?  If so the %s log will be cleared.",
  button1 = "Yes",
  button2 = "No",
  timeout = 0,
  whileDead = 1,
  OnAccept = function(self)
    flsys:ClearLog(self.etype)
  end,
}

StaticPopupDialogs["DECAY"] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = "Decay ep and gp?",
  button1 = "Yes",
  button2 = "No",
  timeout = 0,
  hideOnEscape = 1,
  enterClicksFirstButton = true,
  whileDead = 1,
  OnAccept = function()
    flsys:DecayRaid()
  end,
}

StaticPopupDialogs["SENDDATA"] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = "Send data to:",
  button1 = ACCEPT,
  button2 = CANCEL,
  timeout = 0,
  whileDead = 1,
  hideOnEscape = 1,
  hasEditBox = 1,
  maxLetters = 25,
  OnAccept = function(self)
    local name = self.editBox:GetText()
    flsys:SendData(name)
  end
}

StaticPopupDialogs["RECEIVE_DATA"] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = "%s is sending you loot standings. Accept them?\nWARNING: Imported data overwrites stored data!",
  button1 = "Yes",
  button2 = "No",
  timeout = 0,
  whileDead = 1,
  OnAccept = function(self)
    flsys:ImportEntered(self.text, self.player)
  end,
  OnCancel = function(self, reason)
    SendAddonMessage("flamingsys", "datareply:".."declined your data.", "WHISPER", self.player)
  end,
}

StaticPopupDialogs["CHANGE_ILVL"] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = "|cff6600FFflaming loot|r\n\nNew patch! Update your base ilvl to match current content",
  button1 = "Okay",
  button2 = "No thanks",
  OnAccept = function(self)
    flsys:ProcessChatCmd("opt")
  end,
  OnCancel = function(self, reason)
    flsys:Print("GP may become disproportiately high compared to EP if you don't adjust it.  Type |cffffff78/fs opt|r to edit your settings later.")
  end,
}

StaticPopupDialogs["DELETE_LOG_ITEM"] = {
  preferredIndex = STATICPOPUPS_NUMDIALOGS,
  text = "Are you sure you want to delete this award?  It will be removed from the log and the corresponding EP or GP will be removed from the player's total.",
  button1 = ACCEPT,
  button2 = CANCEL,
  OnAccept = function(self)
    flsys:Debug(1, "ALTER LOG", 'delete on accept')
    FLLogItem:Hide()
    flsys:Undo(self.log_i, self.raid_i, true)
  end,
}

